# -*- coding: utf-8 -*-
import re
from collections import namedtuple

DanskebankLine = namedtuple('DanskebankLine', ['date', 'text', 'amount', 'balance', 'status', 'cleared'])
amount_pattern = r'^-?\d{1,3}(\.\d{3})*,\d{2}$'
column_patterns = {'date':    r'^\d{2}\.\d{2}\.\d{4}$',
                   'text':    r'^.+$',
                   'amount':  amount_pattern,
                   'balance': amount_pattern,
                   'status':  r'^Udført$',
                   'cleared': r'^(Ja|Nej)$',
                   }
column_patterns = {column: re.compile(regex) for column, regex in column_patterns.items()}
txn_date_descends = False


def getlines(path):
    import csv
    import datetime
    import locale
    from . import validate_line
    from .ynab import YnabLine

    with open(path, 'r', encoding='iso-8859-1') as handle:
        transactions = csv.reader(handle, delimiter=';', quotechar='"',
                                  quoting=csv.QUOTE_MINIMAL)
        locale.setlocale(locale.LC_ALL, 'da_DK.UTF-8')
        # Skip headers
        next(transactions)
        for raw_line in transactions:
            try:
                line = DanskebankLine(*raw_line)
                validate_line(line, column_patterns)

                date = datetime.datetime.strptime(line.date, '%d.%m.%Y')
                payee, memo = parse_text(line.text)
                category = u''
                amount = locale.atof(line.amount)
                if amount > 0:
                    outflow = 0.0
                    inflow = amount
                else:
                    outflow = -amount
                    inflow = 0.0
            except Exception:
                import sys
                msg = (u"There was a problem on line {line} in {path}\n"
                       .format(line=transactions.line_num, path=path))
                sys.stderr.write(msg)
                raise

            yield YnabLine(date, payee, category, memo, outflow, inflow)


def parse_text(text):
    result = re.match(r'^(?P<payee>.+\S)\s+(?P<trnsnum>\d{5})$', text)
    if result is not None:
        matches = result.groupdict()
        return matches['payee'], 'txn #{trnsnum}'.format(trnsnum=matches['trnsnum'])

    result = re.match(r'^VDK (?P<currency>[A-Z]{3})\s+(?P<amount>(\d{1,3})(\.\d{3})*,\d{2})$', text)
    if result is not None:
        matches = result.groupdict()
        payee = 'VisaDankort {currency}'.format(currency=matches['currency'])
        memo = '{currency} {amount}'.format(currency=matches['currency'], amount=matches['amount'])
        return payee, memo

    return text, u''



#!/usr/bin/env python3


def main():
    import docopt
    from . import __doc__ as doc
    opts = docopt.docopt(doc)
    try:
        if opts['consolidate']:
            consolidate(opts)
        if opts['convert']:
            convert(opts)
    except Exception as e:
        if opts['--debug']:
            raise
        import sys
        sys.stderr.write(str(e) + '\n')
        sys.exit(1)


def consolidate(opts):
    from . import load_formula
    from .formats import ynab
    import os
    from itertools import chain
    formula, formula_module = load_formula(opts['FORMULA'])

    out_prefix = os.path.join(formula['outpath'], formula['outprefix'])

    # Get all files that were converted with this formula
    converted_files = sorted(find_files(out_prefix), reverse=formula_module.txn_date_descends)
    all_converted = [file[0] for file in converted_files]
    if len(all_converted) < 2:
        raise Exception('There must be at least 2 files to consolidate')

    # Get all lines from those converted files
    all_lines = list(chain(*(ynab.getlines(path) for path in all_converted)))

    if len(all_lines) > 0:
        output_filepath, fromdate, todate, increment = get_filename_parts(out_prefix, all_lines)

        # Write import lines to outputfile
        with ynab.write_file(output_filepath) as put_line:
            for line in all_lines:
                put_line(line)

        print("Wrote {written} transactions to {path}"
              .format(written=len(all_lines),
                      read=len(all_lines),
                      path=output_filepath))
    else:
        print("No transactions found for formula `{}'".format(opts['FORMULA']))

    # Remove all files that have been consolidated
    for converted in all_converted:
        os.remove(converted)


def convert(opts):
    from . import load_formula
    from .formats import ynab
    import os.path
    import shutil
    from itertools import chain
    formula, formula_module = load_formula(opts['FORMULA'])

    out_prefix = os.path.join(formula['outpath'], formula['outprefix'])
    archive_prefix = os.path.join(formula['archivepath'], formula['outprefix'])

    for import_file in opts['INFILE']:
        # Get the lines to import
        # (and cache the list, we will iterate through it multiple times)
        all_lines = list(formula_module.getlines(import_file))

        # Find the daterange we are importing
        import_min = min([line.date for line in all_lines])
        import_max = max([line.date for line in all_lines])

        # Find previously converted files
        previously_converted = list(find_daterange(out_prefix, import_min, import_max))

        # Get all lines from those previously converted files
        prev_lines = chain(*(ynab.getlines(path) for path in previously_converted))

        # Filter previous lines to only contain transactions inside the import daterange
        # (also cache this one, we do multiple lookups)
        consolidation_lines = list([line for line in prev_lines if import_min <= line.date <= import_max])

        # Multiply import lines by "factor"
        all_lines = list([factor_line(line, formula.get('factor', 1)) for line in all_lines])

        # Filter import lines using the previously converted transactions
        unique_lines = [line for line in all_lines if line not in consolidation_lines]

        if len(unique_lines) > 0:
            output_filepath, fromdate, todate, increment = get_filename_parts(out_prefix, unique_lines)

            # Write import lines to outputfile
            with ynab.write_file(output_filepath) as put_line:
                for line in unique_lines:
                    put_line(line)

            print("Wrote {written} out of {read} transactions to {path}"
                  .format(written=len(unique_lines),
                          read=len(all_lines),
                          path=output_filepath))
        else:
            print("No unique lines found in {path}".format(path=import_file))

        # Archive original file (use the total daterange for the filename)
        date_suffix = '-' + import_max.strftime('%Y%m%d') + '-' + import_min.strftime('%Y%m%d')
        archive_filepath = archive_prefix + date_suffix + '.csv'
        increment = 0
        while os.path.exists(archive_filepath):
            increment += 1
            archive_filepath = archive_prefix + date_suffix + '-' + str(increment) + '.csv'
        shutil.move(import_file, archive_filepath)


def factor_line(line, factor):
    from .formats.ynab import YnabLine
    return YnabLine(line.date, line.payee, line.category, line.memo,
                    round(factor * line.outflow, 2), round(factor * line.inflow, 2))


def find_daterange(prefix, min_date, max_date):
    for path, fromdate, todate, increment in find_files(prefix):
        if fromdate <= max_date and todate >= min_date:
            yield path


def find_files(prefix):
    import datetime
    import re
    import glob
    file_pattern = re.compile(
        r'^' + re.escape(prefix) + r'\-(?P<to>\d{8})\-(?P<from>\d{8})(\-(?P<inc>\d+))?\.csv$')
    for path in glob.glob(prefix + '-' + '[0-9]' * 8 + '-' + '[0-9]' * 8 + '*.csv'):
        result = file_pattern.match(path)
        if result is None:
            raise Exception('Found file that does not match pattern: ' + path)
        fromdate = datetime.datetime.strptime(result.group('from'), '%Y%m%d')
        todate = datetime.datetime.strptime(result.group('to'), '%Y%m%d')
        increment = None
        if result.group('inc') is not None:
            increment = int(result.group('inc'))
        yield path, fromdate, todate, increment


def get_filename_parts(out_prefix, lines):
    import os.path
    # Find the daterange in the lines supplied and create the name for the output file with that
    # We reverse the date order so that the files are sorted by newest transaction
    fromdate = min([line.date for line in lines])
    todate = max([line.date for line in lines])
    date_suffix = '-' + todate.strftime('%Y%m%d') + '-' + fromdate.strftime('%Y%m%d')
    output_filepath = out_prefix + date_suffix + '.csv'
    # Handle duplicate filenames by tacking on an increment counter
    increment = 0
    while os.path.exists(output_filepath):
        increment += 1
        output_filepath = out_prefix + date_suffix + '-' + str(increment) + '.csv'
    return output_filepath, fromdate, todate, increment


if __name__ == '__main__':
    main()





def validate_line(line, column_patterns):
    for name, regex in column_patterns.items():
        value = getattr(line, name)
        if regex.match(value) is None:
            msg = ("Column `{column}' with value `{value}' did not match pattern {pattern}"
                   .format(column=name, value=value, pattern=regex.pattern))
            raise Exception(msg)



"""
An Eika, or SDC portalbank, module for attempting some smart extraction

This module was primarily created for customers of a bank that is part of Eika's cooperation, which uses the danish
SDC "portalbank" solution. Exports from all banks from there should follow the same conventions, but your mileage may
vary.

The module makes a lot of assumptions, and as such should be used on a very, small set of transactions initially.
Once you've established some trust, and confirmed that my assumptions also hold for use-case, then feel free to help
tidy up or come with suggestions.

This module is to be considered untested, unstable, alpha, et cetera.

.. module:: eika

.. moduleauthor:: Thor K. Høgås <thor at roht no>


"""

import re
from collections import namedtuple

EikaLine = namedtuple('EikaLine', ['date', 'txndate', 'text', 'amount', 'balance'])
amount_pattern = r'^-?\d{1,3}(\.\d{3})*,\d{2}$'
date_pattern = r'^\d{2}\.\d{2}\.\d{4}'
text_date_pattern = r'\d{4}-\d{2}-\d{2}$'
column_patterns = {'date':    date_pattern,
                   'txndate': date_pattern,
                   'text':    r'^.+$',
                   'amount':  amount_pattern,
                   'balance': amount_pattern,
                   }
column_patterns = {column: re.compile(regex) for column, regex in column_patterns.items()}


def getlines(path):
    import csv
    import datetime
    import locale
    from . import validate_line
    from .ynab import YnabLine

    with open(path, 'r', encoding='utf-8-sig') as handle:
        transactions = csv.reader(handle, delimiter=';', quotechar='"',
                                  quoting=csv.QUOTE_ALL)
        locale.setlocale(locale.LC_ALL, 'nb_NO.UTF-8')

        for raw_line in transactions:
            try:
                line = EikaLine(*raw_line)
                validate_line(line, column_patterns)

                date = datetime.datetime.strptime(line.date, '%d.%m.%Y')
                payee = line.text
                category = ''
                memo = ''

                # Extra cool parsing, because of this shitty export
                if re.match(r'^Varekjøp', line.text):
                    payee = re.search(r'(?<=Varekjøp ).+?(?= betal dato)', line.text)[0]
                    date = datetime.datetime.strptime(re.search(text_date_pattern, line.text)[0], '%Y-%m-%d')

                elif re.match(r'^VISA VARE', line.text):
                    m = re.search(r'(?<=VISA VARE \w{16} )(?P<date>\d{2}.\d{2})  ?(?P<cost>\w{0,3} ?\d*?,?\d*?) (?P<payee>.+?) (?=Kurs)', line.text)
                    m = m if m is not None \
                        else re.search(r'(?<=VISA VARE \w{16} )(?P<date>\d{2}.\d{2})  ?(?P<cost>\w{0,3} ?\d*?,?\d*?) (?P<payee>.+)', line.text)
                    m = m if m is not None \
                        else re.search(r'(?<=VISA VARE \w{16} )(?P<date>\d{2}.\d{2})  ?(?P<payee>.+)', line.text)
                    transaction = m.groupdict()
                    payee = transaction.get('payee', line.text)
                    memo = transaction.get('cost', '') if transaction.get('cost') != '0,00' else ''

                elif re.match(r'Lønn', line.text):
                    payee = re.search(r'(?<=Lønn - ).*', line.text)[0]

                else:
                    payee = line.text

                amount = locale.atof(line.amount.replace('.', ''))
                if amount > 0:
                    outflow = 0.0
                    inflow = amount
                else:
                    outflow = -amount
                    inflow = 0.0
            except Exception as e:
                import sys
                msg = ("There was a problem on line {line} in {path}, Python line {line_code}\n"
                       .format(line=transactions.line_num, path=path, line_code=sys.exc_info()[2].tb_lineno))
                sys.stderr.write(raw_line[2] + "\n")
                sys.stderr.write(msg)
                raise e

            yield YnabLine(date, payee, category, memo, outflow, inflow)



# -*- coding: utf-8 -*-
import re
from collections import namedtuple

SaxoTraderGoLine = namedtuple('SaxoTraderGoLine', [
    'account_id',
    'posting_date',
    'value_date',
    'product',
    'net_change',
    'cash_balance'
])
date_pattern = r'^[0123]\d-([01]\d|[a-z]{3})-[12]\d{3}$'
amount_pattern = r'^-?\d+(\.\d{1,2})?$'
# 'account_id': r'^\d{6}INET$',
column_patterns = {'posting_date': date_pattern,
                   'value_date': date_pattern,
                   # 'product': ,
                   'net_change': amount_pattern,
                   'cash_balance': amount_pattern,
                   }
column_patterns = {column: re.compile(regex) for column, regex in column_patterns.items()}
txn_date_descends = True


def getlines(path):
    import csv
    import datetime
    import locale
    from . import validate_line
    from .ynab import YnabLine

    with open(path, 'r', encoding='utf-8') as handle:
        transactions = csv.reader(handle, delimiter=',', quotechar='"',
                                  quoting=csv.QUOTE_MINIMAL)
        locale.setlocale(locale.LC_ALL, 'en_US.UTF-8')
        # Skip headers
        next(transactions)
        for raw_line in transactions:
            try:
                line = SaxoTraderGoLine(*raw_line)
                validate_line(line, column_patterns)

                try:
                    date = datetime.datetime.strptime(line.value_date, '%d-%m-%Y')
                except ValueError:
                    date = datetime.datetime.strptime(line.value_date, '%d-%b-%Y')
                payee, memo = parse_text(line.product)
                category = u''
                amount = locale.atof(line.net_change)
                if amount > 0:
                    outflow = 0.0
                    inflow = amount
                else:
                    outflow = -amount
                    inflow = 0.0
            except Exception:
                import sys
                msg = (u"There was a problem on line {line} in {path}\n"
                       .format(line=transactions.line_num, path=path))
                sys.stderr.write(msg)
                raise

            yield YnabLine(date, payee, category, memo, outflow, inflow)


def parse_text(text):
    result = re.match(r'^(?P<payee>.+) (?P<txnid>\d{9,})$', text)
    if result is not None:
        matches = result.groupdict()
        return matches['payee'], '{payee} txn #{txnid}'.format(payee=matches['payee'], txnid=matches['txnid'])
    else:
        return text, u''



import re
from collections import namedtuple

BankNordikLine = namedtuple('BankNordikLine', ['date', 'txndate', 'text', 'amount', 'balance', 'unknown', 'unknown2'])

amount_pattern = r'^-?\d{1,3}(\.\d{3})*,\d{2}$'
date_pattern = r'^\d{2}-\d{2}-\d{4}$'
column_patterns = {'date':    date_pattern,
                   'text':    r'^.+$',
                   'amount':  amount_pattern,
                   'balance': amount_pattern,
                   }
column_patterns = {column: re.compile(regex) for column, regex in column_patterns.items()}
txn_date_descends = True


def getlines(path):
    import csv
    import datetime
    import locale
    from . import validate_line
    from .ynab import YnabLine

    with open(path, 'r', encoding='utf-8-sig') as handle:
        transactions = csv.reader(handle, delimiter=';', quotechar='"',
                                  quoting=csv.QUOTE_ALL)
        locale.setlocale(locale.LC_ALL, 'da_DK.UTF-8')

        for raw_line in transactions:
            try:
                line = BankNordikLine(*raw_line)
                validate_line(line, column_patterns)

                date = datetime.datetime.strptime(line.date, '%d-%m-%Y')
                payee = line.text
                category = ''
                memo = ''
                amount = locale.atof(line.amount)
                if amount > 0:
                    outflow = 0.0
                    inflow = amount
                else:
                    outflow = -amount
                    inflow = 0.0
            except Exception:
                import sys
                msg = ("There was a problem on line {line} in {path}\n"
                       .format(line=transactions.line_num, path=path))
                sys.stderr.write(msg)
                raise

            yield YnabLine(date, payee, category, memo, outflow, inflow)



import re
from collections import namedtuple

NordnetLine = namedtuple('NordnetLine', ['line_id', 'bogf_date', 'trns_date', 'val_date', 'trns_type',
                                         'stock_name', 'instr_type', 'isin', 'quantity', 'price',
                                         'interest', 'fee', 'amount', 'currency', 'buy_price',
                                         'result', 'total_qty', 'saldo', 'exch_rate', 'trns_text',
                                         'shred_date', 'verification_number',
                                         ])
date_pattern = r'^\d{4}-\d{2}-\d{2}$'
optional_date_pattern = r'^(\d{4}-\d{2}-\d{2}|)$'
amount_pattern = r'^-?\d{1,3}(\.\d{3})*,\d{2,}$'
pos_amount_pattern = r'^\d{1,3}(\.\d{3})*,\d{2,}$'
column_patterns = {'line_id':     r'^\d{9,}$',
                   'bogf_date':   date_pattern,
                   'trns_date':   date_pattern,
                   'val_date':    date_pattern,
                   'trns_type':   r'^.+$',
                   'stock_name':  r'^.*$',
                   'instr_type':  r'^.*$',
                   'isin':        r'^.*$',
                   'quantity':    r'^(\d+|)$',
                   'price':       pos_amount_pattern,
                   'interest':    pos_amount_pattern,
                   'fee':         pos_amount_pattern,
                   'amount':      amount_pattern,
                   'currency':    r'^[A-Z]{3}$',
                   'buy_price':   pos_amount_pattern,
                   'result':      amount_pattern,
                   'total_qty':   amount_pattern,
                   'saldo':       amount_pattern,
                   'exch_rate':   pos_amount_pattern,
                   'trns_text':   r'^.*$',
                   'shred_date':  optional_date_pattern,
                   'verification_number': r'^\d{8,}$',
                   }
column_patterns = {column: re.compile(regex) for column, regex in column_patterns.items()}
txn_date_descends = True


def getlines(path):
    import csv
    import datetime
    import locale
    from . import validate_line
    from .ynab import YnabLine

    with open(path, 'r', encoding='iso-8859-1') as handle:
        transactions = csv.reader(handle, delimiter=';', quotechar='"',
                                  quoting=csv.QUOTE_MINIMAL)
        locale.setlocale(locale.LC_ALL, 'da_DK.UTF-8')
        conv = {'EUR': 7.46, 'USD': 6.68, 'DKK': 1}
        # Skip headers
        next(transactions)
        for raw_line in transactions:
            if len(raw_line) == 0:
                # export may contain empty lines
                continue
            try:
                line = NordnetLine(*raw_line)
                validate_line(line, column_patterns)

                date = datetime.datetime.strptime(line.bogf_date, '%Y-%m-%d')
                if line.instr_type == 'Aktie':
                    payee = '{action} {code}'.format(action=line.trns_type.capitalize(), code=line.stock_name)
                    memo = '{qty} stk. til {cur} {price}'.format(qty=line.quantity, price=line.price, cur=line.currency)
                else:
                    payee = line.trns_type.capitalize()
                    memo = ''
                category = ''
                amount = round(locale.atof(line.amount) * conv[line.currency], 2)
                if amount > 0:
                    outflow = 0.0
                    inflow = amount
                else:
                    outflow = -amount
                    inflow = 0.0
            except Exception:
                import sys
                msg = ("There was a problem on line {line} in {path}\n"
                       .format(line=transactions.line_num, path=path))
                sys.stderr.write(msg)
                raise

            yield YnabLine(date, payee, category, memo, outflow, inflow)



# -*- coding: utf-8 -*-
import re
from collections import namedtuple

AmericanExpressLine = namedtuple('AmericanExpressLine', ['date', 'payee', 'payer',
                                                         'account_no', 'amount'])
date_pattern = r'^\d{2}/\d{2}/\d{4}$'
column_patterns = {'date':       date_pattern,
                   'payee':      r'^.*$',
                   'payer':      r'^.*$',
                   'account_no': r'^-?\d+$',
                   'amount':     r'^-?\d+,\d{2}$',
                   }
column_patterns = {column: re.compile(regex) for column, regex in column_patterns.items()}
txn_date_descends = True


def getlines(path):
    import csv
    import datetime
    import locale
    from . import validate_line
    from .ynab import YnabLine

    with open(path, 'r', encoding='utf-8') as handle:
        transactions = csv.reader(handle, delimiter=',', quotechar='"',
                                  quoting=csv.QUOTE_MINIMAL)
        locale.setlocale(locale.LC_ALL, 'de_DE.UTF-8')

        # Skip headers
        next(transactions)
        for raw_line in transactions:
            try:
                line = AmericanExpressLine(*raw_line)
                validate_line(line, column_patterns)

                date = datetime.datetime.strptime(line.date, '%d/%m/%Y')
                payee = line.payee.capitalize()
                payee = re.sub(r'  .*$', '', payee)
                category = ''
                memo = line.payer.title()
                amount = locale.atof(line.amount)
                if amount > 0:
                    outflow = 0.0
                    inflow = -amount
                else:
                    outflow = amount
                    inflow = 0.0
            except Exception:
                import sys
                msg = ("There was a problem on line {line} in {path}\n"
                       .format(line=transactions.line_num, path=path))
                sys.stderr.write(msg)
                raise

            yield YnabLine(date, payee, category, memo, outflow, inflow)



"""ynab-csv-converter
convert: Convert transactionlists from various bank formats to the nYNAB format
consolidate: Combine all converted files from a formula into one single file

Usage:
  ynab-csv-converter [options] convert FORMULA INFILE...
  ynab-csv-converter [options] consolidate FORMULA

Options:
  -h, --help   show this help
  -d, --debug  print debug messages (e.g. include stacktrace in errors)

Supported formats:
  nordnet              nordnet.dk ("XLS compatible")
  sparekassen          Sparekassen Kronjylland
  sparekassen-erhverv  Sparekassen Kronjylland business account
  danskebank           Danske Bank (tab separated)
  hypovereinsbank      HypoVereinsbank
  americanexpress      American Express
  starmoney            Starmoney 13
  laanspar             Lån & Spar
  eika                 Eika
  saxotradergo         Excel export from SaxoTrader Go, converted to CSV
  banknordik           BankNordik
"""

import os.path


__version__ = '1.0.1'


def load_yaml(path):
    import yaml
    with open(path, 'r') as stream:
        return yaml.safe_load(stream)


def validate_formula(formula):
    import jsonschema
    schema_path = os.path.normpath(os.path.join(os.path.dirname(__file__), 'formula-schema.yml'))

    schema = load_yaml(schema_path)
    try:
        jsonschema.validate(formula, schema)
    except jsonschema.ValidationError as e:
        raise FormulaError(e.message, formula, e.path)


def load_formula(formula_path):
    from . import load_yaml
    from . import validate_formula
    import importlib
    formula = load_yaml(formula_path)
    validate_formula(formula)
    module_name = 'ynab_csv_converter.formats.' + formula['format']
    module = importlib.import_module(module_name)
    return formula, module


class FormulaError(Exception):
    def __init__(self, message, settings_path, data_path=None):
        super(FormulaError, self).__init__(message)
        self.message = message
        self.settings_path = settings_path
        self.data_path = data_path

    def __str__(self):
        if self.data_path is not None:
            path = '.'.join(map(str, self.data_path))
            return ('{msg}\n  File path: {file}\n  Data path: {datapath}'
                    .format(msg=self.message, file=self.settings_path, datapath=path))
        return '{file}: {msg}'.format(msg=self.message, file=self.settings_path)



# -*- coding: utf-8 -*-
import re
from collections import namedtuple

HypoVereinsbankLine = namedtuple('HypoVereinsbankLine', ['account_no', 'date', 'txn_date',
                                                         'payee1', 'payee2', 'text',
                                                         'amount', 'currency'])
date_pattern = r'^\d{2}\.\d{2}\.\d{4}$'
column_patterns = {'account_no': r'^\d+$',
                   'date':       date_pattern,
                   'txn_date':   date_pattern,
                   'payee1':     r'^.*$',
                   'payee2':     r'^.*$',
                   'text':       r'^.*$',
                   'amount':     r'^-?\d{1,3}(\.\d{3})*,\d{2}$',
                   'currency':   r'^[A-Z]{3}$',
                   }
column_patterns = {column: re.compile(regex) for column, regex in column_patterns.items()}
txn_date_descends = False


def getlines(path):
    import csv
    import datetime
    import locale
    from . import validate_line
    from .ynab import YnabLine

    with open(path, 'r', encoding='utf-16-le') as handle:
        transactions = csv.reader(handle, delimiter=';', quotechar='"',
                                  quoting=csv.QUOTE_MINIMAL)
        locale.setlocale(locale.LC_ALL, 'de_DE.UTF-8')

        # Skip headers
        next(transactions)
        for raw_line in transactions:
            try:
                line = HypoVereinsbankLine(*raw_line)
                validate_line(line, column_patterns)

                date = datetime.datetime.strptime(line.date, '%d.%m.%Y')
                payee = line.payee1.capitalize()
                category = ''
                memo = ''
                if len(payee) == 0:
                    payee = line.text.capitalize()
                else:
                    memo = line.text.capitalize()
                if len(line.payee2) > 0:
                    memo += ' Empfänger 2: ' + line.payee2.capitalize()
                amount = locale.atof(line.amount.replace('.', ''))
                if amount > 0:
                    outflow = 0.0
                    inflow = amount
                else:
                    outflow = -amount
                    inflow = 0.0
            except Exception:
                import sys
                msg = ("There was a problem on line {line} in {path}\n"
                       .format(line=transactions.line_num, path=path))
                sys.stderr.write(msg)
                raise

            yield YnabLine(date, payee, category, memo, outflow, inflow)



from contextlib import contextmanager
import re
from collections import namedtuple

YnabLine = namedtuple('YnabLine', ['date', 'payee', 'category', 'memo', 'outflow', 'inflow'])
amount_pattern = r'^\d+\.\d{2}$'
column_patterns = {'date':     r'^\d{2}/\d{2}/\d{4}$',
                   'payee':    r'^.+$',
                   'category': r'^.*$',
                   'memo':     r'^.*$',
                   'outflow':  amount_pattern,
                   'inflow':   amount_pattern,
                   }
column_patterns = {column: re.compile(regex) for column, regex in column_patterns.items()}


def getlines(path):
    import csv
    import datetime
    from . import validate_line

    with open(path, 'r', encoding='utf-8') as handle:
        transactions = csv.reader(handle, delimiter=',', quotechar='"',
                                  quoting=csv.QUOTE_MINIMAL)
        # Skip headers
        next(transactions)
        for raw_line in transactions:
            try:
                line = YnabLine(*raw_line)
                validate_line(line, column_patterns)

                date = datetime.datetime.strptime(line.date, '%d/%m/%Y')
                outflow = float(line.outflow)
                inflow = float(line.inflow)
            except Exception:
                import sys
                msg = ("There was a problem on line {line} in {path}\n"
                       .format(line=transactions.line_num, path=path))
                sys.stderr.write(msg)
                raise
            yield YnabLine(date, line.payee, line.category, line.memo, outflow, inflow)


@contextmanager
def write_file(path):
    import csv
    with open(path, 'w', encoding='utf-8') as handle:
        transactions_out = csv.writer(handle, delimiter=',', quotechar='"',
                                      quoting=csv.QUOTE_MINIMAL)
        transactions_out.writerow(['Date', 'Payee', 'Category', 'Memo', 'Outflow', 'Inflow'])

        def putline(line):
            transactions_out.writerow([line.date.strftime('%d/%m/%Y'),
                                       line.payee,
                                       line.category,
                                       line.memo,
                                       format(line.outflow, '.2f'),
                                       format(line.inflow, '.2f'),
                                       ])
        yield putline
